from netLoader.RemoteControl import ConnectionRemote, ConnectionRemotePackage


class AdapterConnectionRemote(ConnectionRemote):
    """
    For modify the method 'get_remote_f' should
    override method '_get_remote_f' and '__exit__', '__enter__' method
    you can save the IO object(example save the process buffer object) in 'self._session'
    or implement you own protocol.
    """
    pass


class AdapterConnectionRemotePackage(ConnectionRemotePackage):
    """
    For modify the method 'get_remote_package' should
    override method '_get_remote_package' and '__exit__', '__enter__' method
    you can save the IO object(example save the process buffer object) in 'self._session'
    or implement you own protocol.
    """
    pass




"""
reference:
        https://peps.python.org/pep-0302/
        https://python3-cookbook.readthedocs.io/zh-cn/latest/c10/p11_load_modules_from_remote_machine_by_hooks.html
"""
import sys

from netLoader.PathFinder import RemotePathFinder

from Utils.LinkDealer import LinksDealer

_addr_path_cache = {}

_installed = False


def handle_addr(path):
    if LinksDealer.filter_links(path):
        if path in _addr_path_cache:
            finder = _addr_path_cache[path]
        else:
            finder = RemotePathFinder(path)
            _addr_path_cache[path] = finder
        return finder
    else:
        pass


def register_remote_pkg_server(url):
    if _installed:
        sys.path.append(url)


def install_path_hook():
    sys.path_hooks.append(handle_addr)
    sys.path_importer_cache.clear()
    global _installed
    _installed = True


def remove_path_hook():
    if _installed:
        sys.path_hooks.remove(handle_addr)
        sys.path_importer_cache.clear()



import imp
import importlib
import sys
import requests

from netLoader.RemoteControl import ConnectionRemotePackage


class RemoteModuleLoader(importlib.abc.SourceLoader):
    def __init__(self, addr):
        super().__init__()
        self._base_addr = addr
        self._source_cache = {}

    def module_repr(self, module):
        return f"<RemoteModule {module.__name__} from {module.__file__}>"

    # Override 'load_module' method
    def load_module(self, fullname):
        code = self.get_code(fullname)
        mod = sys.modules.setdefault(fullname, imp.new_module(fullname))
        mod.__file__ = self.get_filename(fullname)
        mod.__loader__ = self
        mod.__package__ = fullname.rpartition(".")[0]

    def get_data(self, path):
        ...

    def get_filename(self, fullname):
        return self._base_addr + '/' + fullname.split('.')[-1] + ".py"

    def get_code(self, fullname):
        src = self.get_source(fullname)
        return compile(src, self.get_filename(fullname), 'exec')

    def get_source(self, fullname):
        filename = self.get_filename(fullname)

        if filename in self._source_cache:
            return self._source_cache[filename]
        try:
            with ConnectionRemotePackage() as conn:
                remote_pkg_content = conn.get_remote_package(filename).text
                self._source_cache[filename] = remote_pkg_content
                return remote_pkg_content
        except (ConnectionError, requests.ConnectionError, requests.HTTPError) as _:
            raise ImportError(f"Can not load {filename}")

    def is_package(self, fullname):
        return False



from netLoader.ModuleLoader import RemoteModuleLoader


class RemotePackageLoader(RemoteModuleLoader):
    def load_module(self, fullname) -> None:
        mod = super().load_module(fullname)
        mod.__path__ = [self._base_addr]
        mod.__package__ = fullname

    def get_filename(self, fullname) -> str:
        return self._base_addr + "/" + '__init.py'

    def is_package(self, fullname) -> bool:
        return True

    


import requests
import types


class BaseConnection(object):
    def __init__(self): ...

    def __enter__(self):
        raise NotImplementedError("Subclasses must implement __enter__")

    def __exit__(self, exc_type, exc_val,*args, **kwargs):
        raise NotImplementedError("Subclasses must implement __exit__")

    def get_remote_f(self, remote_addr, *args, **kwargs):
        raise NotImplementedError("Subclasses must implement 'get_remote_f'.")

    def get_remote_package(self, remote_addr, *args, **kwargs): ...


class ConnectionRemote(BaseConnection):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._session: requests.Session = None

    def __enter__(self):
        self._session = requests.Session()
        return self

    def __exit__(self, exc_type, exc_val, *args):
        self._session.close()

    def _get_remote_f(self, remote_addr, *args, **kwargs):
        resp = self._session.get(remote_addr)
        return resp

    def get_remote_f(self, remote_addr, *args, **kwargs):
        result = self._get_remote_f(remote_addr, *args, **kwargs)
        return result

    def get_remote_package(self, remote_addr, *args, **kwargs): ...


class ConnectionRemotePackage(BaseConnection):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._session: requests.Session = None

    def __enter__(self):
        self._session = requests.Session()

    def __exit__(self, exc_type, exc_val, *args, **kwargs):
        self._session.close()

    def get_remote_f(self, remote_addr, *args, **kwargs): ...

    def _get_remote_package(self, remote_addr, *args, **kwargs) -> requests.Response:
        result = self._session.get(remote_addr, *args, **kwargs)
        return result

    def get_remote_package(self, remote_addr, *args, **kwargs) -> requests.Response:
        return self._get_remote_package(remote_addr, *args, **kwargs)



import importlib.abc
from netLoader.ModuleLoader import RemoteModuleLoader
from Utils.LinkDealer import LinksDealer
from netLoader.PackageLoader import RemoteModuleLoader, RemotePackageLoader


class RemotePathFinder(importlib.abc.PathEntryFinder):
    def __init__(self, base_addr):
        self._links = None
        self._loader = RemoteModuleLoader(base_addr)
        self._base_addr = base_addr

    def find_module(self, fullname):
        parts = fullname.split('.')
        base_name = parts[-1]

        # 检查 link 缓存 如果为空则用LinkDealer 去获取links
        if self._links is None:
            self._links = []
            self._links = LinksDealer.get_link(self._base_addr)

        # 检查是否 package
        if base_name in self._links:
            full_addr = self._base_addr + '/' + base_name
            loader = RemotePackageLoader(full_addr)
            try:
                loader.load_module(fullname)
            except ImportError as e:
                loader = None
            return loader, [full_addr]

        # 检查是否为 module
        remote_filename = base_name + '.py'
        if remote_filename in self._links:
            return self._loader, []
        return None, []

    def invalidate_caches(self):
        self._links = None

